'''A simple script to use with a Statestep model treated as a
decision table, in which some variables are treated as inputs
and others (zero or more) as outputs.
A suitable model with which to try this is activitiesOutcomes
(one of the sample models supplied with Statestep): in that
model, the inputs are 'Temperature', 'Weather' and 'Wind' and
there is one output variable: 'Activity'.
'''

import sys


def decision():
   assert model.vars(), \
   'Model needs at least one variable with one or more values.'
   assert model.events(), 'Need a model with at least one event.'
   inputvars = getinputvars()
   outputvars = [x for x in model.vars() if x not in inputvars]
   while True:
      print
      print 'Enter the value of each input variable. You'
      print 'can press Enter to accept the [default value]'
      print 'or \'?\' to list the possible values.'
      comb = model.Combination()
      for var in inputvars:
         val = None
         while not val:
            vals = model.vals(var)
            default = vals[0]
            val = raw_input('Enter the value of %s [%s]: ' \
            % (var, default)).strip()
            if not val:
               val = default
            elif val not in vals:
               print 'The possible values are:', ' '.join(vals)
               val = None
         comb = comb.copy({var: val})
      print 'Input:', getvalues(comb, inputvars)
      print '(This input is combination number %d of %d.)' % \
      (comb.num(inputvars) + 1, model.Combination.count(inputvars))
      default = model.events()[0]
      if len(model.events()) > 1:
         event = None
         while not event:
            event = raw_input('Enter the event [%s]: ' \
            %  default).strip()
            if not event:
               event = default
            elif event not in model.events():
               print 'The possible events are:', \
               ' '.join(model.events())
               event = None
      else:
         event = default
      rule, outcome = model.applyRules(event, comb)
      if rule:
         print 'Rule', rule, 'applies.'
         if outputvars:
            print 'Output:', getvalues(outcome, outputvars)
      else:
         print 'No rule applies for this case.'
      while True:
         again = raw_input('Try again ([y]/n)? ').strip()
         if not again or again[0].lower() == 'y':
            break
         elif again[0].lower() == 'n':
            print
            return


def getvalues(comb, varlist):
   '''Returns a string giving the values of particular variables
   in the specified Combination.'''
   return ', '.join(['%s=%s'%(var,comb[var]) for var in varlist])


def getinputvars():
   '''Find out which variables are supposed to be the inputs
   to the decision table.'''
   inputs = []
   while not inputs:
      print 'The variables are:', ' '.join(model.vars())
      inputs = raw_input('Which are the input variables? ')
      inputs = inputs.replace(',',' ')
      inputs = inputs.split()
      if not inputs:
         print 'There must be at least one input variable.'
      else:
         unknowns = [x for x in inputs if x not in model.vars()]
         if unknowns:
            print '"%s" is not a variable, please try again.' % \
            unknowns[0]
            inputs = []
   return inputs


if __name__ == '__main__':
   if len(sys.argv) > 2:
      sys.exit('No more than one argument expected.');
   if len(sys.argv) == 2:
      mfile = sys.argv[1].split('.py')[0]
      try:
         model = __import__(mfile)
      except Exception:
         sys.exit("Could not import '%s'." % mfile)
   else:
      import model
   decision()
