'''A simple script to use with a Statestep model treated as a
finite state machine.
A suitable model with which to try this is harelsWatch (one
of the sample models that comes with Statestep).
'''

import sys


def fsm():
   assert model.vars(), \
   'Model needs at least one variable with one or more values.'
   assert model.events(), 'Need a model with at least one event.'
   # try and pick a state that is allowed by the constraints
   for state in model.CombIter():
      if model.valid(state):
         break;
   print 'Initial state:'
   printstate(state)
   while True:
      print 'This state is combination number %d.' % state.num()
      if not model.valid(state):
         print 'Note: this state is not valid (according to',
         print 'constraint "%s").' % model.explainInvalid(state)
      while True:
         event = raw_input('Enter an event ("." to exit): ').strip()
         print
         if event in model.events():
            break
         elif event == '.':
            sys.exit(0)
         print '"%s" is not an event. The events are:' % event
         print ', '.join(model.events())
      print 'Event "%s" occurs ...' % event
      if not model.expected(event, state):
         print 'Note: this event was not expected in this state',
         print '(see constraint "%s" for event "%s").' % \
         (model.explainUnexpected(event, state), event)
      rule, next = model.applyRules(event, state)
      if rule:
         print 'Rule "%s" applies; the new state is:' % rule
         state = next
      else:
         print 'No rule applies. The state is unchanged:'
      printstate(state)


def printstate(state):
   for var in model.vars():
      print '   %20s = %s' % (var, state[var])


if __name__ == '__main__':
   if len(sys.argv) > 2:
      sys.exit('No more than one argument expected.');
   if len(sys.argv) == 2:
      mfile = sys.argv[1].split('.py')[0]
      try:
         model = __import__(mfile)
      except Exception:
         sys.exit("Could not import '%s'." % mfile)
   else:
      import model
   fsm()
