<#ftl strip_whitespace = true>
<#-- Variables without any values must be ignored. -->
<#if doc.model.variables?size != 0>
   <#assign usable_variables =
   doc.model.variables["variable[./value]"]
   />
<#else>
   <#assign usable_variables = []/>
</#if>
'''Code generated from a Statestep model.

The principal methods and classes are:

   vars() - get the variables in the model.

   vals() - get the possible values of a variable.

   Combination - a mapping from variables to values.

   events() - get the list of events.

   applyRules() - for a given case (event + combination), find the
      first rule that applies and get the outcome of that rule.

Among several other methods, valid() can be used to find out
whether a particular combination is allowed by the constraints;
if it isn't, explainInvalid() can tell you why.

Iterators: The API also includes optimized iterator classes and
methods which may be useful in contexts such as testing.
'''

# Code generated using the Statestep Python Template.
# Template copyright (c) 2006, Michael Breen.
# Open Source. Licensed under the Apache License, Version 2.0.
# See template package for more details.


_vars = [
<#list usable_variables as var>
'${var.@name}',
</#list>
]


_varvals = {
<#list usable_variables as var>
'${var.@name}': [
   <#list var[".//value[not(./value)]"] as leaf>
   '<#rt>
      <#list leaf?ancestors('value')?reverse as branch>
         ${branch.@name}-<#t>
      </#list>
         ${leaf.@name}'<#t>
      <#if !leaf_has_next>
         , ]<#t>
      </#if>
,
   </#list>
</#list>
}


# shortcut to get the number of values of a variable given its index
_numvals = [len(_varvals[_v]) for _v in _vars]


def vars():
   '''vars() -> tuple of strings

   Returns all the variables (in model definition order).
   '''
   return tuple(_vars)


def vals(var):
   '''vals(varname) -> tuple of strings

   Returns the values of the specified variable.

   Default values:
   The first value listed is always the variable's
   'default value'. This is an arbitrary value assigned to
   the variable wherever its value might otherwise be
   undefined; external code should not depend on values
   being listed in any particular order.
   '''
   return tuple(_varvals[var])


class _Vals(object):
   '''An immutable set of values of a variable.

   The following code illustrates usage, assuming there is
   a variable 'Light' with values 'on', 'off' and 'broken'.

      light_off = _Vals('Light','off')
      light_off.var == 'Light'
      light_off[0] == 'off'
      light_all = _Vals('Light','*')
      [x for x in light_all] == ['broken', 'off', 'on']
      len(light_all) == 3
      light_all == light_off | _Vals('Light','broken','on')
      light_all - light_off == _Vals('Light','on','broken')
      light_off & _Vals('Light','broken','on') == _Vals('Light')

   Treated as a sequence, the values in a _Vals objects will
   always be listed in the same order as they appear in the
   list returned by vals().
   '''
   def __init__(self, varname, *valnames):
      self._var = str(varname)
      try:
         allvals = _varvals[self._var]
      except KeyError:
         raise Exception, 'No such variable: "%s".' % self._var
      self._vals = []
      for name in [str(x) for x in valnames]:
         if name.endswith('*'):
            vlist = [v for v in allvals if v.startswith(name[:-1])]
         else:
            vlist = [v for v in [name] if v in allvals]
         if not vlist:
            raise Exception, '"%s" not a value of variable "%s".' \
            % (name, self._var)
         self._vals = [v for v in allvals if v in self._vals \
         or v in vlist]
   def __str__(self):
      return self._var + '={' + '|'.join(self._vals) + '}'
   def __and__(self, other):
      '''Intersection.'''
      self._checkOperand(other)
      result = _Vals(self._var)
      result._vals = [v for v in self._vals if v in other._vals]
      return result
   def __or__(self, other):
      '''Union.'''
      self._checkOperand(other)
      result = _Vals(self._var)
      result._vals = [v for v in _varvals[self._var] if \
      v in self._vals or v in other._vals]
      return result
   def __sub__(self, other):
      '''Set difference.'''
      self._checkOperand(other)
      result = _Vals(self._var)
      result._vals = [v for v in self._vals if v not in other._vals]
      return result
   def _checkOperand(self,other):
      if not isinstance(other, _Vals):
         raise TypeError, \
         "Both operands need to be of type '_Vals'."
      if self._var != other._var:
         raise ValueError, \
         "The operands must be values of the same variable."
   var = property(lambda self: self._var, doc \
   = 'The variable to which this set of values relates.')
   def __getitem__(self, key):
      return self._vals[key]
   def __len__(self):
      return self._vals.__len__()
   def __eq__(self, other):
      return isinstance(other, _Vals) and self._var == other._var \
      and self._vals == other._vals
   def __ne__(self, other):
      return not self.__eq__(other)


class _CombinationMask(object):
   '''Associates a set of allowed values with each variable.

   This defines a set of allowed combinations: those in which
   each variable has one of the allowed values.
   '''
   _defaults = dict([(_x, _Vals(_x, '*')) for _x in _vars])
   def __init__(self, *initializers):
      self._vals = { }
      for i in initializers:
         self._vals[i.var] = i
   def __getitem__(self, key):
      '''Returns the values allowed for the specified variable.'''
      if self._vals.has_key(key):
         return self._vals[key]
      else:
         return _CombinationMask._defaults[key]
   def __contains__(self,item):
      '''Checks if the specified Combination is allowed.'''
      if not isinstance(item, Combination):
         raise TypeError
      for var, vals in self._vals.items():
         if item[var] not in vals:
            return False
      return True
   def copy(self, *initializers):
      '''Returns a modified copy of this mask.'''
      return _CombinationMask(\
      *([v[1] for v in self._vals.items()] + list(initializers)))
   def combination(self):
      '''Returns the first combination allowed by this mask.'''
      return Combination(dict([(k, v[0]) for k, v \
      in self._vals.items()]))
   def __str__(self):
      return '[' + ', '.join([str(self[k]) for k in _vars]) + ']'


class _TargetRow(object):
   '''Corresponds to the second row of table cells in a
   tabular rule.
   '''
   def __init__(self, *initializers):
      self._vals = { }
      for i in initializers:
         self._vals[i.var] = i
   def transform(self, state):
      '''Returns a mask of allowed next states (outcomes)
      based on a current state.
      '''
      return _CombinationMask( \
      *([_Vals(x, state[x]) for x in _vars \
      if not self._vals.has_key(x)] \
      + [x[1] for x in self._vals.items()]))
   def __str__(self):
      return '+'.join([str(self._vals[x]) for x in self._vals])


class Combination(object):
   '''Maps each variable to one of its values.

   In the context of a finite state machine model, a mapping
   of this kind may define the current state.

   Combinations are immutable; however, a modified copy of
   a Combination can be obtained using the copy() method.

   Note on methods count(), get(), num():
   These accept a list of variables as an optional argument.
   In that case, these methods number Combinations according
   to the values of the listed variables only; in effect, it is
   as if the unlisted variables did not exist (in the case of
   get(), unlisted variables are given their default values in
   the returned Combination).
   '''
   def __init__(self, initializer={ }):
      '''Combination([dict])

      The argument is a mapping from variables to values.
      Example:

         comb = Combination({'Door':'closed', 'Light':'on'})

      Any variable for which no value is specified is given its
      default value.
      '''
      self.__offsets = [0] * len(_vars)
      self.__update(initializer)
   def copy(self, initializer):
      '''Returns a modified copy of this combination.

      The argument is a mapping specifying any values to be
      overridden. For example, combNew below will be the same
      as comb, except for the value of 'Door':

         combNew = comb.copy({'Door':'open'})
      '''
      comb = Combination()
      comb.__offsets = self.__offsets[:]
      comb.__update(initializer)
      return comb
   def __update(self, initializer):
      for key, val in initializer.items():
         iv = _vars.index(key)
         offset = _varvals[key].index(val)
         self.__offsets[iv] = offset
   def count(varlist = None):
      '''Returns the total number of distinct combinations
      that can be formed. This is simply the product of the
      number of values each variable can have.

      Optional argument varlist: see note above.
      '''
      # (avoiding product(), reduce() for compatibility)
      result = 1
      for iv in Combination._rvindices(varlist):
         result *= _numvals[iv]
      return result
   count = staticmethod(count)
   def get(n, varlist = None):
      '''Returns combination number 'n', where 0 <= n < count().

      Combination number 0 is that in which all variables have
      their default values.

      Optional argument varlist: see note above.
      '''
      comb = Combination()
      for iv in Combination._rvindices(varlist):
         n, comb.__offsets[iv] = divmod(n, _numvals[iv])
      if n:
         raise ValueError
      return comb
   get = staticmethod(get)
   _revrange = tuple(range(len(_vars))[::-1])
   def _rvindices(varlist):
      '''Returns the indices of variables in _vars, reverse sorted.

      In other words, the index of the least significant variable
      (with respect to the numbering of Combinations) comes first.
      For example, if _vars were ('A','B','C','D') and ['B','A','D']
      were passed to this function then it would return [3,1,0].

      An argument of None is treated as a list of all variables.
      '''
      if varlist is None:
         return Combination._revrange
      else:
         ivars = [False] * len(_vars)
         for v in varlist:
            ivars[_vars.index(v)] = True
         return [x for x in Combination._revrange if ivars[x]]
   _rvindices = staticmethod(_rvindices)
   def num(self, varlist = None):
      '''Returns a number uniquely identifying this combination.
      This will be in the range 0 .. (Combination.count() - 1).

      Optional argument varlist: see note above.
      '''
      result = 0
      factor = 1
      for iv in Combination._rvindices(varlist):
         result += factor * self.__offsets[iv]
         factor *= _numvals[iv]
      return result
   def _iterator(self, varlist = None):
      '''Returns an iterator over Combinations.

      The first Combination returned by the iterator is this one.
      After that, for successive combinations A, B returned by
      the iterator, B.num() > A.num().

      If no argument is passed then B.num() = A.num() + 1
      and the number of the last Combination returned is
      Combination.count() - 1.

      If an argument is passed then B.num() is the lowest
      number greater than A.num() for which the values of the
      specified variables are different from their values in
      A *and* every other Combination already returned.
      '''
      # The algorithm is explained at the end of the file.
      yield self
      indices = Combination._rvindices(varlist)
      if not indices:
         return
      iterations = 1
      # index into indices referencing the most significant
      # variable in that list which has been incremented so far
      imax = 0
      # number of different possible combinations of values of
      # the variables referenced by indices[imax:]
      cycles = _numvals[indices[imax]]
      offsets = self.__offsets[:]
      for i in range(indices[imax] + 1, len(offsets)):
         offsets[i] = 0
      while True:
         if iterations < cycles:
            for i in range(imax + 1):
               # index of most significant variable incremented
               msvi = indices[i]
               offsets[msvi] += 1
               if offsets[msvi] < _numvals[msvi]:
                  break
               offsets[msvi] = 0
         else:
            while not iterations < cycles:
               imax += 1
               try:
                  msvi = indices[imax]
               except IndexError:
                  return
               cycles *= _numvals[msvi]
            for i in range(msvi + 1, len(offsets)):
               offsets[i] = 0
            offsets[msvi] += 1
            if not offsets[msvi] < _numvals[msvi]:
               offsets[msvi] = 0
         if offsets[msvi] == 0:
            while msvi:
               msvi -= 1
               offsets[msvi] += 1
               if offsets[msvi] < _numvals[msvi]:
                  break
               offsets[msvi] = 0
            else:
               return
            while imax + 1 < len(indices) \
            and indices[imax + 1] >= msvi:
               imax += 1
               cycles *= _numvals[indices[imax]]
         comb = Combination()
         comb.__offsets = offsets[:]
         yield comb
         iterations += 1
   def _mask(self):
      '''Return a _CombinationMask allowing only this Combination.'''
      return _CombinationMask(*[_Vals(k, self[k]) for k in _vars])
   def __hash__(self):
      # num() might conceivably return a long int
      return hash(self.num())
   def __eq__(self, other):
      return isinstance(other, Combination) and \
      self.__offsets == other.__offsets
   def __ne__(self, other):
      return not self.__eq__(other)
   def __cmp__(self, other):
      '''Comparisons (<, <=, >=, >) between Combinations are
      equivalent to comparing the results of their num() methods.
      For example: (comb1 > comb2) == (comb1.num() > comb2.num())
      '''
      try:
         return self.num() - other.num()
      except AttributeError:
         raise NotImplementedError
   def __getitem__(self, key):
      '''[variable] -> value

      Returns the value of the specified variable. Example:

         if comb['Door'] == 'open':
            print 'The door is open.'
      '''
      return _varvals[key][self.__offsets[_vars.index(key)]]
   def __str__(self):
      return '[' + ', '.join([self[k] for k in _vars]) + ']'


def valid(comb):
   '''Check if a combination (state) is possible.

   Returns True if the constraints in the model (the global
   constraints, including the constraint table) allow the
   specified combination, otherwise False.
   '''
   return not explainInvalid(comb)


def explainInvalid(comb):
   '''explainInvalid(Combination) -> string

   Returns a string identifying one of the model's constraints,
   according to which the combination should not be possible.
   Returns None if the combination / state is possible.
   '''
   ifthen = _explainInvalidFromTable(comb)
   if ifthen:
      explanation = 'Constraint Table: %s => %s' % ifthen
   else:
      explanation = explainUnexpected(None, comb)
   return explanation


def _explainInvalidFromTable(comb):
   '''_explainInvalidFromTable(Combination) -> (ifvals, thenvals)
   Returns a tuple of constraining and constrained values from
   the constraint table identifying a constraint which makes
   the argument invalid.
   Returns None if there is no such constraint.
   '''
   for ifvals in _constraintTable.keys():
      if comb[ifvals.var] in ifvals:
         for thenvals in _constraintTable[ifvals]:
            if not comb[thenvals.var] in thenvals:
               return ifvals, thenvals


def events():
   '''events() -> tuple of strings

   Returns all the events in the model.
   '''
   return _events


def expected(event, state):
   '''Check whether an event is expected in a given state.

   Returns False if there is a constraint for the specified
   event which excludes the state, otherwise True.

   NOTE: This method does not check that the state is valid
   (that is, it ignores the global constraints); a call to
   valid() would normally precede a call to this method.
   '''
   return not explainUnexpected(event, state)


def explainUnexpected(event, state):
   '''Find out why an event is unexpected.

   Returns a string identifying the constraint stating
   that the event should never occur in the specified
   state; returns None if the event is not unexpected.
   '''
   # If 'event' is None then this method searches the
   # global constraints.
   for con, formula in _constraints[event].items():
      if not _evalConstraintFormula(formula, state):
         return con


class CombIter(object):
   '''Iterator over all Combinations which satisfy some condition.

   Combinations are always iterated over in numerical order,
   that is, for any Combinations c1 and c2 returned in that
   order, c1.num() < c2.num().

   In this, the base class, there is no condition; subclasses
   provide for different kinds of condition. These generally
   execute much faster than code using a separate test outside
   the iterator. For example, the following code iterates over
   all valid (possible) Combinations:

      # slow
      for s in CombIter():
         if valid(s):
            ...

   However, the CombIter returned by validCombinationsIter()
   provides a better alternative:

      # fast
      for s in validCombinationsIter()
         ...

   For a large model with many constraints, the performance
   improvement can be quite dramatic. Of course, for still
   larger models, even the latter loop will become slow.
   '''
   def __init__(self):
      self.reset(Combination())
   def reset(self, comb):
      '''Make the iterator rewind or skip forward.

      The next Combination returned by the iterator will be
      either the argument or, if comb does not satisfy the
      iterator's condition, the lowest-numbered Combination
      following it that does.
      '''
      self._itr = comb._iterator()
   def __iter__(self):
      return self
   def next(self):
      return self._itr.next()


class MaskIter(CombIter):
   '''Iterator over Combinations in which some values are fixed.

   For example:

      for c in MaskIter(['A', 'C', 'D']):
         ...

   The argument is a list (or other container) of variables
   whose values can change; other variables have their default
   values. Thus, if there are two other variables, 'B' and
   'E', then the code above would loop over every Combination
   in which 'B' and 'E' had their default values.

   An optional second argument can be used to specify non-default
   values for the fixed variables. This can be a Combination or
   simply a mapping from variables to values. For example, the
   following object iterates over all Combinations in which the
   value of variable 'B' is 'b4' (since variable 'A' is listed
   in the first argument, its value in the second is ignored):

      MaskIter(['A','C','D','E'], Combination({'A':'a2','B':'b4'}))
   '''
   def __init__(self, varlist, fixed = None):
      self._varlist = varlist
      if fixed:
         self._fixed = dict( \
         [(v, fixed[v]) for v in _vars if v not in varlist])
      else:
         self._fixed = { }
      self._num = 0
   def reset(self, comb):
      dct = { }
      carry = False
      for var in _vars[::-1]:
         if var not in self._varlist:
            ival = _varvals[var].index(comb[var])
            try:
               ifix = _varvals[var].index(self._fixed[var])
            except KeyError:
               ifix = 0
            if ival < ifix:
               carry = False
               dct = { }
            elif ival > ifix:
               carry = True
               dct = { }
         else:
            if carry:
               ival = _varvals[var].index(comb[var])
               try:
                  dct[var] = _varvals[var][ival + 1]
                  carry = False
               except IndexError:
                  pass
            else:
               dct[var] = comb[var]
      if carry:
         # overflow; cause StopIteration to be raised in next()
         self._num = Combination.count(self._varlist)
      else:
         self._num = Combination(dct).num(self._varlist)
   def next(self):
      try:
         comb = Combination.get(self._num, self._varlist)
      except ValueError:
         raise StopIteration
      if self._fixed:
         comb = comb.copy(self._fixed)
      self._num += 1
      return comb


class GangedIter(CombIter):
   '''An iterator efficiently combining two or more CombIters.

   A ganged iterator effectively 'AND's the conditions of its
   sources. For example, the following object iterates over
   every Combination in which variables other than 'A' and 'C'
   have their default values and which is also valid:

      GangedIter(MaskIter(['A','C']), validCombinationsIter())

   NOTE: The arguments are assumed to be independent of each other
   (e.g., the same iterator object must not be passed twice) and
   they should not used anywhere else. For example, calling next()
   on any of the arguments even after the GangedIter has been
   created may cause it to skip.
   '''
   def __init__(self, *combIters):
      self._iters = combIters
      if len(combIters) < 2:
         raise TypeError, 'Expected at least 2 CombIter arguments.'
      self._itr = self._getiter()
   def reset(self, comb):
      for it in self._iters:
         it.reset(comb)
      # (in case generator code has terminated via StopIteration:)
      self._itr = self._getiter()
   def _getiter(self):
      nextc = { }
      while True:
         for it in self._iters:
            nextc[it] = it.next()
         combs = nextc.values()
         maxc = max(combs)
         minc = min(combs)
         while not maxc == minc:
            minc = maxc
            for it in self._iters:
               if nextc[it] < maxc:
                  it.reset(maxc)
                  maxc = nextc[it] = it.next()
         yield maxc


class _AllowedByTable(CombIter):
   '''Iterator over Combinations allowed by the constraint table.
   '''
   def __init__(self):
      CombIter.__init__(self)
   def next(self):
      while True:
         comb = CombIter.next(self)
         ifthen = _explainInvalidFromTable(comb)
         if ifthen:
            subitr = comb._iterator([x.var for x in ifthen])
            subitr.next()
            while True:
               comb = subitr.next()
               if comb[ifthen[1].var] in ifthen[1] \
               or not comb[ifthen[0].var] in ifthen[0]:
                  break
            self.reset(comb)
         else:
            return comb


class _AllowedByFormulas(CombIter):
   '''CombIter over Combinations allowed by a set of constraints.
   '''
   def __init__(self, event):
      '''The event identifies the set of constraint formulas to be
      checked; for the global constraints, this argument is None.
      '''
      CombIter.__init__(self)
      self._event = event
   def next(self):
      while True:
         comb = CombIter.next(self)
         con = explainUnexpected(self._event, comb)
         if con:
            subitr = comb._iterator(_constraintVars[self._event][con])
            subitr.next()
            self.reset(subitr.next())
         else:
            return comb


def validCombinationsIter():
   '''validCombinationsIter() -> CombIter

   Returns an iterator over the possible Combinations, that is,
   the Combinations allowed by the constraint table and the
   global constraint formulas.
   '''
   return GangedIter(_AllowedByTable(), _AllowedByFormulas(None))


def validExpectedIter(event):
   '''validExpectedIter(event) -> CombIter

   Returns a CombIter over the valid states in which a
   particular event is expected. The returned iterator
   provides an efficient alternative to code like:

      for state in validCombinationsIter():
         if expected(event, state):
            ...

   Note, however, that the following code, which iterates just
   once over the valid combinations, will typically run much
   faster than the alternative of using multiple iterators
   returned by this method:

      for state in validCombinationsIter():
         for event in events():
            if expected(event, state):
               ...
   '''
   return GangedIter(_AllowedByTable(), _AllowedByFormulas(None), \
   _AllowedByFormulas(event))


def applyRules(event, currState, extraLogic = None):
   '''applyRules(event, currentState) -> (rule, nextState)

   Returns the first rule that applies when event occurs in
   currentState, together with the next state (outcome) as
   determined by the rule; returns (None, None) if no rule
   exists for the specified case.

   The last parameter, 'extraLogic', provides an optional
   callback mechanism. Arguments should have the signature:

   extraLogic(event, rule, current, next) -> Combination | None

   Whenever applyRules finds a rule that applies, it passes
   this to extraLogic, together with a next state / outcome
   consistent with that rule. If extraLogic returns None or
   False then applyRules looks for another rule instead.
   Otherwise, it returns the outcome returned by extraLogic.
   In effect, the default extraLogic is:

      def extraLogic(event, rule, current, next):
         return next

   (For some examples, see the HTML documentation accompanying
   the template package from which this code was generated.)
   '''
   for rule, logic in _rulesTable[event].items():
      if currState in logic[0]:
         mask = logic[1].transform(currState)
         if _ruleFormulas[event].has_key(rule):
            formula = _ruleFormulas[event][rule]
            nextState = _evalRuleFormula(formula, currState, mask)
            if nextState == None:
               continue
         else:
            nextState = mask.combination()
         if extraLogic:
            nextState = extraLogic(event, rule, currState, nextState)
            if not nextState:
               continue
         return rule, nextState
   return None, None


def ruleApplies(event, rule, currState, nextState):
   '''Check whether a rule allows a particular outcome.

   Returns True if the specified rule for the specified event
   allows a transition from currState to nextState, otherwise
   False.

   One possible use of this function is where 'extra logic'
   helps to determine the outcome of a (non-deterministic)
   rule: in this case, it could be used to double-check that
   the outcome provided by the extralogic is one allowed by
   the rule, for example

      rule, outcome = applyRules(event, comb, untrustedLogic)
      if rule:
         assert ruleApplies(event, rule, comb, outcome)
         ...
   '''
   logic = _rulesTable[event][rule]
   if currState not in logic[0]:
      return False
   if nextState not in logic[1].transform(currState):
      return False
   if _ruleFormulas[event].has_key(rule):
      formula = _ruleFormulas[event][rule]
      if not _evalRuleFormula(formula, currState, \
      nextState._mask()):
         return False
   return True


def _evalConstraintFormula(formula, state):
   '''Returns True if the formula allows the specified state,
   otherwise False.
   '''
   return _evalNode(formula, state)


def _evalRuleFormula(formula, current, nextMask):
   '''Returns the state allowed by nextMask to which the formula
   maps the current state. If the formula and the mask allow for
   more than one such state then it chooses one arbitrarily.
   Returns None if there is no such state.
   '''
   mask = _evalNode(formula, current, nextMask)
   if not mask:
      return None
   else:
      return mask.combination()


class _RuleSolved(Exception):
   '''Raised to indicate a rule formula has been solved.
   The single argument is a _CombinationMask of target values
   (defining possible next states) which satisfy the formula
   given a specified set of source values (current state).
   '''
   pass


def _evalNode(formula, current, nextMask = True, node = None):
   '''With respect to a given current state, determine whether
   a constraint formula holds or a rule formula can be satisfied.
   For a constraint, the first two arguments are used and True
   is returned if the formula holds in state current.
   For a rule formula, nextMask is specified and limits the
   possible next states (target states) which can be used to
   satisfy the rule. If the formula can be satisfied then
   a _CombinationMask is returned specifying a (perhaps
   non-unique) set of possible next states which would
   satisfy the formula: this will either be nextMask or a
   further restriction of it.
   False is returned for any formula that cannot be satisfied.
   '''
   if formula == None:
      return nextMask
   if node == None:
      node = formula
   try:
      if node.type == 'source':
         smask = node.child[0]
         if current[smask.var] in smask:
            return nextMask
         else:
            return False
      # target values can occur only in rule formulas
      elif node.type == 'target':
         tmask = node.child[0]
         # an empty set of target values signifies 'no change'
         if not tmask:
            tmask = _Vals(tmask.var, current[tmask.var])
         restricted = nextMask[tmask.var] & tmask
         excluded = nextMask[tmask.var] - tmask
         if not restricted:
            return False
         elif excluded:
            # it is possible the overall formula will be satisfied
            # only if this clause is satisfied - or only if it is
            # not satisfied; recurse to allow for both possibilities
            result = _evalNode(formula, current, \
            nextMask.copy(restricted))
            if not result:
               result =  _evalNode(formula, current, \
               nextMask.copy(excluded))
            if result:
               raise _RuleSolved(result)
            else:
               return result
         else:
            return nextMask
      elif node.type == 'not':
         if _evalNode(formula, current, nextMask, node.child[0]):
            return False
         else:
            return nextMask
      elif node.type == 'and':
         for subnode in node.child:
            nextMask = _evalNode(formula, current, nextMask, subnode)
            if not nextMask:
               return False
         return nextMask
      elif node.type == 'or':
         for subnode in node.child:
            newMask = _evalNode(formula, current, nextMask, subnode)
            if newMask:
               return newMask
         return False
      elif node.type == 'implies':
         lhs = _evalNode(formula, current, nextMask, node.child[0])
         if lhs:
            return _evalNode(formula, current, lhs, node.child[1])
         else:
            return nextMask
      else:
         raise RuntimeError, 'Unknown parse node.'
   except _RuleSolved, soln:
      # This exception is used not only for efficiency but also
      # to smuggle an altered nextMask through any 'not' nodes.
      if node == formula:
         return soln.args[0]
      else:
         raise


class _pnode(object):
   '''Parse node in a constraint formula or a rule formula.'''
   def __init__(self, type, *nodes):
      self.type = type
      self.child = nodes


_events = (
<#list doc.model.events.event as event>
'${event.@name}',
</#list>
)


_constraintTable = {
<#list doc.model.simple_constraints.constraint_group as row>
   <#if (row[
"./value_set[./@type='constrained'][not(./@content='dont_care')]"]
   ?size > 0)>
${values_to_code(row["./value_set[./@type='constraining']"][0])}: (
      <#list
row["./value_set[./@type='constrained'][not(./@content='dont_care')]"]
      as vals>
   ${values_to_code(vals)}<#rt>
         <#if !vals_has_next>
            , )<#t>
         </#if>
,
      </#list>
   </#if>
</#list>
}


# constraints; variables used in the constraints
_constraints = { }
_constraintVars = { }
# global constraints:
_constraints[None] = { }
_constraintVars[None] = { }
<#list doc.model.constraints.constraint as constraint>
   <@constraint_formula constraint/>
   <@constraint_vars constraint/>
</#list>
# event constraints:
<#list doc.model.events.event as event>
_constraints['${event.@name}'] = { }
_constraintVars['${event.@name}'] = { }
   <#list event.constraints.constraint as constraint>
      <@constraint_formula constraint, "'${event.@name}'"/>
      <@constraint_vars constraint, "'${event.@name}'"/>
   </#list>
</#list>


_rulesTable = {
<#list doc.model.events.event as event>
'${event.@name}': {
   <#list event.rules.rule as rule>
   '${rule.@name}': (
      _CombinationMask(<#rt>
      <#list rule[
"./values_array/value_set[@type='source'][not(@content='dont_care')]"]
      as source_vals>

         ${values_to_code(source_vals)}<#rt>
         <#if source_vals_has_next>
            ,<#t>
         </#if>
      </#list>
),
      _TargetRow(<#rt>
      <#list targetvals(rule) as arg>

         ${arg}<#rt>
         <#if arg_has_next>
            ,<#t>
         </#if>
      </#list>
)),
   </#list>
   },
</#list>
}


_ruleFormulas = { }
<#list doc.model.events.event as event>
_ruleFormulas['${event.@name}'] = { }
   <#list event.rules.rule as rule>
      <#if rule.formula?size != 0>
         _ruleFormulas['${event.@name}']['${rule.@name}'] = <#t>
         <#if rule.formula.parse_node?size != 0>
            \<#t>
            <@parse_node_to_code node=rule.formula.parse_node/>

         <#else>
None
         </#if>
      </#if>
   </#list>
</#list>

# Note on Combination._iterator()
# -------------------------------
# The algorithm is key to the speed of validCombinationsIter().
# For anyone who might need to understand it, here is an example.
# Suppose:
#  - there is a variable 'A' with 3 values ('a0', 'a1', 'a2')
#    ('a0' being the default value) and similarly for 4 other
#    variables 'B' .. 'E'
#  - the iterator is obtained from, and thus begins with,
#    Combination({'A':'a1','B':'b1','C':'c1','D':'d1','E':'e1'})
#  - the varlist passed to _iterator() is ['A','C','D']
#    (columns marked '**' below)
# Then the generator progresses as follows:
#
# iterations cycles combination (c)      imax-> c.num(varlist) c.num()
#                    **      **  **
#                   [a1, b1, c1, d1, e1]      D             13     121
#          1      3 [a1, b1, c1, d2, e0]      D             14     123
#          2      9 [a1, b1, c2, d0, e0]      C             15     126
#          3      9 [a1, b1, c2, d1, e0]      C             16     129
#          4      9 [a1, b1, c2, d2, e0]      C             17     132
#          5      9 [a1, b2, c0, d0, e0]      C              9     135
#          6      9 [a1, b2, c0, d1, e0]      C             10     138
#          7      9 [a1, b2, c0, d2, e0]      C             11     141
#          8      9 [a1, b2, c1, d0, e0]      C             12     144
#          9     27 [a2, b0, c0, d0, e0]      A             18     162
#         10     27 [a2, b0, c0, d1, e0]      A             19     165
#         11     27 [a2, b0, c0, d2, e0]      A             20     168
#         12     27 [a2, b0, c1, d0, e0]      A             21     171
#         13     27 [a2, b0, c1, d1, e0]      A             22     174
#         14     27 [a2, b0, c1, d2, e0]      A             23     177
#         15     27 [a2, b0, c2, d0, e0]      A             24     180
#         16     27 [a2, b0, c2, d1, e0]      A             25     183
#         17     27 [a2, b0, c2, d2, e0]      A             26     186
#
# * After 2 iterations, there is a carry to variable 'C'.
# * After 5 iterations, there is a carry to variable 'B' - even
#   though that is not in varlist.
# * After 9 iterations, all possible combinations of variables
#   'C' and 'D' have been returned, so 'A' gets incremented.
# * 18 out of a maximum 27 (=3*3*3) combinations are returned
#   (since no combination with A=a0 is included).
<#-----------------  Macros and Functions ----------------->

<#-- Output code for a constraint formula. -->
<#macro constraint_formula constraint parent="None">
   _constraints[${parent}]['${constraint.@name}'] = <#t>
   <#if constraint.formula.parse_node?size != 0>
      \<#t>
      <@parse_node_to_code node=constraint.formula.parse_node/>

   <#else>
None
   </#if>
</#macro>


<#-- Generate the code for a formula parse tree. -->
<#macro parse_node_to_code node indent="   ">

${indent}_pnode(<#rt>
   <#if node.@type == "values">
      <#local vals = node.value_set/>
      '${vals.@type}', ${values_to_code(vals)})<#t>
   <#else>
      '${node.@type}',<#t>
      <#list node.parse_node as subnode>
         <@parse_node_to_code node=subnode indent=indent + "   "/>
         <#if subnode_has_next>
            ,<#t>
         <#else>
            )<#t>
         </#if>
      </#list>
   </#if>
</#macro>


<#-- Output the list of variables used in a constraint. -->
<#macro constraint_vars constraint parent="None">
   _constraintVars[${parent}]['${constraint.@name}'] = (<#t>
   <#local var_nodes = constraint[".//value_set/@variable"]/>
   <#list usable_variables as var>
      <#if var_nodes?seq_contains(var.@name)>
         '${var.@name}', <#t>
      </#if>
   </#list>
)
</#macro>


<#--
Given a rule node, return a list of the code objects
needed to initialize the corresponding _TargetRow
object in the generated code.
-->
<#function targetvals rule>
   <#local arglist = []/>
   <#list usable_variables as var>
      <#attempt>
         <#--
         this works only if values are specified for 'var',
         otherwise the recover block is executed...
         -->
         <#local vals = rule[
"./values_array/value_set[@type='target'][@variable='${var.@name}']"]
         [0]/>
         <#--
         no object is needed for "no change" variables
         since it is the default
         -->
         <#if vals.@content != "no_change">
            <#local arglist = arglist + ["${values_to_code(vals)}"]/>
         </#if>
      <#recover>
         <#--
         if no values are listed for a variable in the
         model file the default semantics is "don't care"
         -->
         <#local arglist = arglist + ["_Vals('${var.@name}','*')"] />
      </#recover>
   </#list>
   <#return arglist/>
</#function>


<#--
Convert a set of variable values to the corresponding code.
"No change" value sets produce code for an empty set of values.
-->
<#function values_to_code valset>
   <#local code = "_Vals('${valset.@variable}'"/>
   <#switch valset.@content>
      <#case "no_change">
         <#break>
      <#case "dont_care">
         <#local code = code + ",'*'"/>
         <#break>
      <#case "list">
         <#local code = code + ",'"/>
         <#local children=valset.value/>
         <#if children?size == 0>
            <#-- an empty list is an implicit "don't care" -->
         <#local code = code + "*"/>
         <#else>
            <#list val_def_nodes(valset) as val>
               <#local code = code + "${value_id(val)}"/>
               <#if val_has_next>
                  <#local code = code + "','"/>
               </#if>
            </#list>
         </#if>
         <#local code = code + "'"/>
         <#break>
   </#switch>
   <#return code + ")"/>
</#function>


<#--
Given a value node in a variable definition tree,
i.e., below a 'variable' element, return the string
identifying that value or group of values.
-->
<#function value_id def_node>
   <#local id = def_node.@name/>
   <#list def_node?ancestors("value") as group>
      <#local id = group.@name + "-" + id/>
   </#list>
   <#if def_node.value?size != 0>
      <#local id = id + "-*"/>
   </#if>
   <#return id/>
</#function>


<#--
Given a set of value nodes in a value set,
return the corresponding nodes in the variable
value definition tree.
-->
<#function val_def_nodes value_set>
   <#local result = []/>
   <#local varName = value_set.@variable/>
   <#local var =
   doc.model.variables["variable[@name='${varName}']"][0]
   />
   <#list value_set[".//value[not(./value)]"] as group>
      <#local def_node = var/>
      <#local path_to_group = group?ancestors("value")?reverse
      + [group]
      />
      <#list path_to_group as component>
         <#local name=component.@name/>
         <#local def_node = def_node["./value[@name='${name}']"][0]/>
      </#list>
      <#local result = result + [def_node]/>
   </#list>
   <#return result>
</#function>
